@@ -28,6 +28,8 @@ THE SOFTWARE.
 struct kernel_route {
     unsigned char prefix[16];
     int plen;
+    unsigned char src_prefix[16];
+    int src_plen; /* no source prefix <=> src_plen == 0 */
     int metric;
     unsigned int ifindex;
     int proto;
@@ -41,6 +43,7 @@ struct kernel_route {
 #define CHANGE_LINK  (1 << 0)
 #define CHANGE_ROUTE (1 << 1)
 #define CHANGE_ADDR  (1 << 2)
+#define CHANGE_RULE  (1 << 3)
 
 #ifndef MAX_IMPORT_TABLES
 #define MAX_IMPORT_TABLES 10
@@ -49,6 +52,9 @@ struct kernel_route {
 extern int export_table, import_tables[MAX_IMPORT_TABLES], import_table_count;
 
 int add_import_table(int table);
+#define SRC_TABLE_NUM 10
+extern int src_table_idx; /* number of the first table */
+extern int src_table_prio; /* first prio range */
 
 int kernel_setup(int setup);
 int kernel_setup_socket(int setup);
@@ -59,7 +65,9 @@ int kernel_interface_ipv4(const char *ifname, int ifindex,
 int kernel_interface_mtu(const char *ifname, int ifindex);
 int kernel_interface_wireless(const char *ifname, int ifindex);
 int kernel_interface_channel(const char *ifname, int ifindex);
+int kernel_disambiguate(int v4);
 int kernel_route(int operation, const unsigned char *dest, unsigned short plen,
+                 const unsigned char *src, unsigned short src_plen,
                  const unsigned char *gate, int ifindex, unsigned int metric,
                  const unsigned char *newgate, int newifindex,
                  unsigned int newmetric);
@@ -70,3 +78,5 @@ int kernel_addresses(char *ifname, int ifindex, int ll,
 int if_eui64(char *ifname, int ifindex, unsigned char *eui);
 int gettime(struct timeval *tv);
 int read_random_bytes(void *buf, size_t len);
+int kernel_older_than(const char *sysname, int version, int sub_version);
+int kernel_has_ipv6_subtrees(void);
