@@ -382,6 +382,18 @@ kernel_interface_channel(const char *ifname, int ifindex)
     return -1;
 }
 
+int
+kernel_disambiguate(int v4)
+{
+    return 0;
+}
+
+int
+kernel_has_ipv6_subtrees(void)
+{
+    return 0;
+}
+
 int
 kernel_route(int operation, const unsigned char *dest, unsigned short plen,
              const unsigned char *src, unsigned short src_plen,
