@@ -39,10 +39,11 @@ THE SOFTWARE.
 #include "resend.h"
 #include "configuration.h"
 #include "local.h"
+#include "disambiguation.h"
 
 struct babel_route **routes = NULL;
 static int route_slots = 0, max_route_slots = 0;
-int kernel_metric = 0;
+int kernel_metric = 0, reflect_kernel_metric = 0;
 int allow_duplicates = -1;
 int diversity_kind = DIVERSITY_NONE;
 int diversity_factor = 256;     /* in units of 1/256 */
@@ -57,6 +58,7 @@ static int two_to_the_one_over_hl = 0; /* 2^(1/hl) * 0x10000 */
 
 static int
 route_compare(const unsigned char *prefix, unsigned char plen,
+              const unsigned char *src_prefix, unsigned char src_plen,
               struct babel_route *route)
 {
     int i = memcmp(prefix, route->src->prefix, 16);
@@ -65,10 +67,23 @@ route_compare(const unsigned char *prefix, unsigned char plen,
 
     if(plen < route->src->plen)
         return -1;
-    else if(plen > route->src->plen)
+    if(plen > route->src->plen)
         return 1;
-    else
-        return 0;
+
+    if(src_plen == 0) {
+        if(route->src->src_plen > 0)
+            return -1;
+    } else {
+        i = memcmp(src_prefix, route->src->src_prefix, 16);
+        if(i != 0)
+            return i;
+        if(src_plen < route->src->src_plen)
+            return -1;
+        if(src_plen > route->src->src_plen)
+            return 1;
+    }
+
+    return 0;
 }
 
 /* Performs binary search, returns -1 in case of failure.  In the latter
@@ -76,6 +91,7 @@ route_compare(const unsigned char *prefix, unsigned char plen,
 
 static int
 find_route_slot(const unsigned char *prefix, unsigned char plen,
+                const unsigned char *src_prefix, unsigned char src_plen,
                 int *new_return)
 {
     int p, m, g, c;
@@ -90,7 +106,7 @@ find_route_slot(const unsigned char *prefix, unsigned char plen,
 
     do {
         m = (p + g) / 2;
-        c = route_compare(prefix, plen, routes[m]);
+        c = route_compare(prefix, plen, src_prefix, src_plen, routes[m]);
         if(c == 0)
             return m;
         else if(c < 0)
@@ -107,10 +123,11 @@ find_route_slot(const unsigned char *prefix, unsigned char plen,
 
 struct babel_route *
 find_route(const unsigned char *prefix, unsigned char plen,
+           const unsigned char *src_prefix, unsigned char src_plen,
            struct neighbour *neigh, const unsigned char *nexthop)
 {
     struct babel_route *route;
-    int i = find_route_slot(prefix, plen, NULL);
+    int i = find_route_slot(prefix, plen, src_prefix, src_plen, NULL);
 
     if(i < 0)
         return NULL;
@@ -127,9 +144,10 @@ find_route(const unsigned char *prefix, unsigned char plen,
 }
 
 struct babel_route *
-find_installed_route(const unsigned char *prefix, unsigned char plen)
+find_installed_route(const unsigned char *prefix, unsigned char plen,
+                     const unsigned char *src_prefix, unsigned char src_plen)
 {
-    int i = find_route_slot(prefix, plen, NULL);
+    int i = find_route_slot(prefix, plen, src_prefix, src_plen, NULL);
 
     if(i >= 0 && routes[i]->installed)
         return routes[i];
@@ -173,7 +191,8 @@ insert_route(struct babel_route *route)
 
     assert(!route->installed);
 
-    i = find_route_slot(route->src->prefix, route->src->plen, &n);
+    i = find_route_slot(route->src->prefix, route->src->plen,
+                        route->src->src_prefix, route->src->src_plen, &n);
 
     if(i < 0) {
         if(route_slots >= max_route_slots)
@@ -214,7 +233,8 @@ flush_route(struct babel_route *route)
         lost = 1;
     }
 
-    i = find_route_slot(route->src->prefix, route->src->plen, NULL);
+    i = find_route_slot(route->src->prefix, route->src->plen,
+                        route->src->src_prefix, route->src->src_plen, NULL);
     assert(i >= 0 && i < route_slots);
 
     local_notify_route(route, LOCAL_FLUSH);
@@ -370,10 +390,17 @@ route_stream_done(struct route_stream *stream)
     free(stream);
 }
 
-static int
+int
 metric_to_kernel(int metric)
 {
-    return metric < INFINITY ? kernel_metric : KERNEL_INFINITY;
+	if(metric >= INFINITY) {
+		return KERNEL_INFINITY;
+	} else if(reflect_kernel_metric) {
+		int r = kernel_metric + metric;
+		return r >= KERNEL_INFINITY ? KERNEL_INFINITY : r;
+	} else {
+		return kernel_metric;
+	}
 }
 
 /* This is used to maintain the invariant that the installed route is at
@@ -406,7 +433,8 @@ install_route(struct babel_route *route)
         fprintf(stderr, "WARNING: installing unfeasible route "
                 "(this shouldn't happen).");
 
-    i = find_route_slot(route->src->prefix, route->src->plen, NULL);
+    i = find_route_slot(route->src->prefix, route->src->plen,
+                        route->src->src_prefix, route->src->src_plen, NULL);
     assert(i >= 0 && i < route_slots);
 
     if(routes[i] != route && routes[i]->installed) {
@@ -415,16 +443,10 @@ install_route(struct babel_route *route)
         return;
     }
 
-    rc = kernel_route(ROUTE_ADD, route->src->prefix, route->src->plen,
-                      route->nexthop,
-                      route->neigh->ifp->ifindex,
-                      metric_to_kernel(route_metric(route)), NULL, 0, 0);
-    if(rc < 0) {
-        int save = errno;
-        perror("kernel_route(ADD)");
-        if(save != EEXIST)
-            return;
-    }
+    rc = kinstall_route(route);
+    if(rc < 0 && errno != EEXIST)
+        return;
+
     route->installed = 1;
     move_installed_route(route, i);
 
@@ -434,19 +456,13 @@ install_route(struct babel_route *route)
 void
 uninstall_route(struct babel_route *route)
 {
-    int rc;
-
     if(!route->installed)
         return;
 
-    rc = kernel_route(ROUTE_FLUSH, route->src->prefix, route->src->plen,
-                      route->nexthop,
-                      route->neigh->ifp->ifindex,
-                      metric_to_kernel(route_metric(route)), NULL, 0, 0);
-    if(rc < 0)
-        perror("kernel_route(FLUSH)");
-
     route->installed = 0;
+
+    kuninstall_route(route);
+
     local_notify_route(route, LOCAL_CHANGE);
 }
 
@@ -471,19 +487,15 @@ switch_routes(struct babel_route *old, struct babel_route *new)
         fprintf(stderr, "WARNING: switching to unfeasible route "
                 "(this shouldn't happen).");
 
-    rc = kernel_route(ROUTE_MODIFY, old->src->prefix, old->src->plen,
-                      old->nexthop, old->neigh->ifp->ifindex,
-                      metric_to_kernel(route_metric(old)),
-                      new->nexthop, new->neigh->ifp->ifindex,
-                      metric_to_kernel(route_metric(new)));
-    if(rc < 0) {
-        perror("kernel_route(MODIFY)");
+    rc = kswitch_routes(old, new);
+    if(rc < 0)
         return;
-    }
 
     old->installed = 0;
     new->installed = 1;
     move_installed_route(new, find_route_slot(new->src->prefix, new->src->plen,
+                                              new->src->src_prefix,
+                                              new->src->src_plen,
                                               NULL));
     local_notify_route(old, LOCAL_CHANGE);
     local_notify_route(new, LOCAL_CHANGE);
@@ -501,15 +513,9 @@ change_route_metric(struct babel_route *route,
 
     if(route->installed && old != new) {
         int rc;
-        rc = kernel_route(ROUTE_MODIFY, route->src->prefix, route->src->plen,
-                          route->nexthop, route->neigh->ifp->ifindex,
-                          old,
-                          route->nexthop, route->neigh->ifp->ifindex,
-                          new);
-        if(rc < 0) {
-            perror("kernel_route(MODIFY metric)");
+        rc = kchange_route_metric(route, refmetric, cost, add);
+        if(rc < 0)
             return;
-        }
     }
 
     /* Update route->smoothed_metric using the old metric. */
@@ -690,11 +696,12 @@ route_acceptable(struct babel_route *route, int feasible,
    that's probably overkill. */
 
 struct babel_route *
-find_best_route(const unsigned char *prefix, unsigned char plen, int feasible,
-                struct neighbour *exclude)
+find_best_route(const unsigned char *prefix, unsigned char plen,
+                const unsigned char *src_prefix, unsigned char src_plen,
+                int feasible, struct neighbour *exclude)
 {
     struct babel_route *route, *r;
-    int i = find_route_slot(prefix, plen, NULL);
+    int i = find_route_slot(prefix, plen, src_prefix, src_plen, NULL);
 
     if(i < 0)
         return NULL;
@@ -734,6 +741,8 @@ update_route_metric(struct babel_route *route)
         struct neighbour *neigh = route->neigh;
         int add_metric = input_filter(route->src->id,
                                       route->src->prefix, route->src->plen,
+                                      route->src->src_prefix,
+                                      route->src->src_plen,
                                       neigh->address,
                                       neigh->ifp->ifindex);
         change_route_metric(route, route->refmetric,
@@ -785,6 +794,7 @@ update_interface_metric(struct interface *ifp)
 struct babel_route *
 update_route(const unsigned char *id,
              const unsigned char *prefix, unsigned char plen,
+             const unsigned char *src_prefix, unsigned char src_plen,
              unsigned short seqno, unsigned short refmetric,
              unsigned short interval,
              struct neighbour *neigh, const unsigned char *nexthop,
@@ -795,7 +805,7 @@ update_route(const unsigned char *id,
     int metric, feasible;
     int add_metric;
     int hold_time = MAX((4 * interval) / 100 + interval / 50, 15);
-
+    int is_v4;
     if(memcmp(id, myid, 8) == 0)
         return NULL;
 
@@ -804,19 +814,30 @@ update_route(const unsigned char *id,
                 format_prefix(prefix, plen), format_address(nexthop));
         return NULL;
     }
+    if(src_plen != 0 && martian_prefix(src_prefix, src_plen)) {
+        fprintf(stderr, "Rejecting martian route to %s from %s through %s.\n",
+                format_prefix(prefix, plen),
+                format_prefix(src_prefix, src_plen), format_eui64(id));
+        return NULL;
+    }
 
-    add_metric = input_filter(id, prefix, plen,
+    is_v4 = v4mapped(prefix);
+    if(src_plen != 0 && is_v4 != v4mapped(src_prefix))
+        return NULL;
+
+
+    add_metric = input_filter(id, prefix, plen, src_prefix, src_plen,
                               neigh->address, neigh->ifp->ifindex);
     if(add_metric >= INFINITY)
         return NULL;
 
-    route = find_route(prefix, plen, neigh, nexthop);
+    route = find_route(prefix, plen, src_prefix, src_plen, neigh, nexthop);
 
     if(route && memcmp(route->src->id, id, 8) == 0)
         /* Avoid scanning the source table. */
         src = route->src;
     else
-        src = find_source(id, prefix, plen, 1, seqno);
+        src = find_source(id, prefix, plen, src_prefix, src_plen, 1, seqno);
 
     if(src == NULL)
         return NULL;
@@ -925,7 +946,9 @@ send_unfeasible_request(struct neighbour *neigh, int force,
                         unsigned short seqno, unsigned short metric,
                         struct source *src)
 {
-    struct babel_route *route = find_installed_route(src->prefix, src->plen);
+    struct babel_route *route = find_installed_route(src->prefix, src->plen,
+                                                     src->src_prefix,
+                                                     src->src_plen);
 
     if(seqno_minus(src->seqno, seqno) > 100) {
         /* Probably a source that lost its seqno.  Let it time-out. */
@@ -934,6 +957,7 @@ send_unfeasible_request(struct neighbour *neigh, int force,
 
     if(force || !route || route_metric(route) >= metric + 512) {
         send_unicast_multihop_request(neigh, src->prefix, src->plen,
+                                      src->src_prefix, src->src_plen,
                                       src->metric >= INFINITY ?
                                       src->seqno :
                                       seqno_plus(src->seqno, 1),
@@ -958,11 +982,14 @@ consider_route(struct babel_route *route)
     if(!route_feasible(route))
         return;
 
-    xroute = find_xroute(route->src->prefix, route->src->plen);
+    xroute = find_xroute(route->src->prefix, route->src->plen,
+                         route->src->src_prefix, route->src->src_plen);
     if(xroute && (allow_duplicates < 0 || xroute->metric >= allow_duplicates))
         return;
 
-    installed = find_installed_route(route->src->prefix, route->src->plen);
+    installed = find_installed_route(route->src->prefix, route->src->plen,
+                                     route->src->src_prefix,
+                                     route->src->src_plen);
 
     if(installed == NULL)
         goto install;
@@ -984,7 +1011,8 @@ consider_route(struct babel_route *route)
     if(installed && route->installed)
         send_triggered_update(route, installed->src, route_metric(installed));
     else
-        send_update(NULL, 1, route->src->prefix, route->src->plen);
+        send_update(NULL, 1, route->src->prefix, route->src->plen,
+                    route->src->src_prefix, route->src->src_plen);
     return;
 }
 
@@ -1033,6 +1061,7 @@ send_triggered_update(struct babel_route *route, struct source *oldsrc,
         /* Route getting significantly worse */
         urgent = 1;
     else if(unsatisfied_request(route->src->prefix, route->src->plen,
+                                route->src->src_prefix, route->src->src_plen,
                                 route->seqno, route->src->id))
         /* Make sure that requests are satisfied speedily */
         urgent = 1;
@@ -1050,19 +1079,23 @@ send_triggered_update(struct babel_route *route, struct source *oldsrc,
         urgent = 0;
 
     if(urgent >= 2)
-        send_update_resend(NULL, route->src->prefix, route->src->plen);
+        send_update_resend(NULL, route->src->prefix, route->src->plen,
+                           route->src->src_prefix, route->src->src_plen);
     else
-        send_update(NULL, urgent, route->src->prefix, route->src->plen);
+        send_update(NULL, urgent, route->src->prefix, route->src->plen,
+                    route->src->src_prefix, route->src->src_plen);
 
     if(oldmetric < INFINITY) {
         if(newmetric >= oldmetric + 512) {
             send_request_resend(NULL, route->src->prefix, route->src->plen,
+                                route->src->src_prefix, route->src->src_plen,
                                 route->src->metric >= INFINITY ?
                                 route->src->seqno :
                                 seqno_plus(route->src->seqno, 1),
                                 route->src->id);
         } else if(newmetric >= oldmetric + 288) {
-            send_request(NULL, route->src->prefix, route->src->plen);
+            send_request(NULL, route->src->prefix, route->src->plen,
+                         route->src->src_prefix, route->src->src_plen);
         }
     }
 }
@@ -1077,7 +1110,9 @@ route_changed(struct babel_route *route,
         struct babel_route *better_route;
         /* Do this unconditionally -- microoptimisation is not worth it. */
         better_route =
-            find_best_route(route->src->prefix, route->src->plen, 1, NULL);
+            find_best_route(route->src->prefix, route->src->plen,
+                            route->src->src_prefix, route->src->src_plen,
+                            1, NULL);
         if(better_route && route_metric(better_route) < route_metric(route))
             consider_route(better_route);
     }
@@ -1097,17 +1132,20 @@ void
 route_lost(struct source *src, unsigned oldmetric)
 {
     struct babel_route *new_route;
-    new_route = find_best_route(src->prefix, src->plen, 1, NULL);
+    new_route = find_best_route(src->prefix, src->plen,
+                                src->src_prefix, src->src_plen, 1, NULL);
     if(new_route) {
         consider_route(new_route);
     } else if(oldmetric < INFINITY) {
         /* Avoid creating a blackhole. */
-        send_update_resend(NULL, src->prefix, src->plen);
+        send_update_resend(NULL, src->prefix, src->plen,
+                           src->src_prefix, src->src_plen);
         /* If the route was usable enough, try to get an alternate one.
            If it was not, we could be dealing with oscillations around
            the value of INFINITY. */
         if(oldmetric <= INFINITY / 2)
             send_request_resend(NULL, src->prefix, src->plen,
+                                src->src_prefix, src->src_plen,
                                 src->metric >= INFINITY ?
                                 src->seqno : seqno_plus(src->seqno, 1),
                                 src->id);
@@ -1140,7 +1178,8 @@ expire_routes(void)
                 if(route_old(r))
                     /* Route about to expire, send a request. */
                     send_unicast_request(r->neigh,
-                                         r->src->prefix, r->src->plen);
+                                         r->src->prefix, r->src->plen,
+                                         r->src->src_prefix, r->src->src_plen);
             }
             r = r->next;
         }
