@@ -47,7 +47,7 @@ struct babel_route {
 struct route_stream;
 
 extern struct babel_route **routes;
-extern int kernel_metric, allow_duplicates;
+extern int kernel_metric, allow_duplicates, reflect_kernel_metric;
 extern int diversity_kind, diversity_factor;
 extern int keep_unfeasible;
 
@@ -70,9 +70,11 @@ route_metric_noninterfering(const struct babel_route *route)
 }
 
 struct babel_route *find_route(const unsigned char *prefix, unsigned char plen,
-                         struct neighbour *neigh, const unsigned char *nexthop);
+                        const unsigned char *src_prefix, unsigned char src_plen,
+                        struct neighbour *neigh, const unsigned char *nexthop);
 struct babel_route *find_installed_route(const unsigned char *prefix,
-                                   unsigned char plen);
+                        unsigned char plen, const unsigned char *src_prefix,
+                        unsigned char src_plen);
 int installed_routes_estimate(void);
 void flush_route(struct babel_route *route);
 void flush_all_routes(void);
@@ -81,6 +83,7 @@ void flush_interface_routes(struct interface *ifp, int v4only);
 struct route_stream *route_stream(int installed);
 struct babel_route *route_stream_next(struct route_stream *stream);
 void route_stream_done(struct route_stream *stream);
+int metric_to_kernel(int metric);
 void install_route(struct babel_route *route);
 void uninstall_route(struct babel_route *route);
 int route_feasible(struct babel_route *route);
@@ -93,6 +96,8 @@ void change_smoothing_half_life(int half_life);
 int route_smoothed_metric(struct babel_route *route);
 struct babel_route *find_best_route(const unsigned char *prefix,
                                     unsigned char plen,
+                                    const unsigned char *src_prefix,
+                                    unsigned char src_plen,
                                     int feasible, struct neighbour *exclude);
 struct babel_route *install_best_route(const unsigned char prefix[16],
                                  unsigned char plen);
@@ -101,6 +106,8 @@ void update_interface_metric(struct interface *ifp);
 void update_route_metric(struct babel_route *route);
 struct babel_route *update_route(const unsigned char *id,
                            const unsigned char *prefix, unsigned char plen,
+                           const unsigned char *src_prefix,
+                           unsigned char src_plen,
                            unsigned short seqno, unsigned short refmetric,
                            unsigned short interval, struct neighbour *neigh,
                            const unsigned char *nexthop,
